;;; Lepton EDA library - Scheme API
;;; Copyright (C) 2020-2021 Lepton EDA Contributors
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

(define-module (lepton ffi)
  #:use-module (system foreign)

  #:export (liblepton
            ;; Helpers.
            true?
            geda-object->pointer
            geda-object->pointer*
            pointer->geda-object

            ;; Foreign functions.
            edascm_is_config
            edascm_is_object
            edascm_is_page

            lepton_object_get_color
            lepton_object_set_color
            lepton_object_get_id
            lepton_object_get_parent
            lepton_object_get_selectable
            lepton_object_set_selectable
            lepton_object_get_type

            lepton_object_is_arc
            lepton_object_is_box
            lepton_object_is_bus
            lepton_object_is_circle
            lepton_object_is_component
            lepton_object_is_line
            lepton_object_is_net
            lepton_object_is_path
            lepton_object_is_picture
            lepton_object_is_pin
            lepton_object_is_text

            lepton_object_calculate_visible_bounds
            lepton_object_copy
            lepton_object_emit_change_notify
            lepton_object_emit_pre_change_notify
            lepton_object_mirror
            lepton_object_rotate
            lepton_object_translate

            lepton_component_object_get_embedded
            lepton_component_object_embed
            lepton_component_object_unembed

            lepton_picture_object_get_embedded
            lepton_picture_object_embed
            lepton_picture_object_unembed

            set_render_placeholders
            colors_count
            g_rc_parse
            lepton_colormap_color_by_id
            lepton_colormap_disable_color
            lepton_colormap_set_color
            lepton_object_page_set_changed
            print_colors_array
            s_attrib_uniq
            s_attrib_add_entry
            s_clib_add_command
            s_clib_add_directory
            s_clib_add_scm
            s_clib_init
            s_toplevel_page_current))

;;; Helper to check if result of C function is TRUE (non-zero).
(define true? (negate zero?))

(define liblepton
  (dynamic-link (or (getenv "LIBLEPTON") "liblepton")))

;;; Brief syntax macro for defining lazy foreign functions.
(define-syntax define-lff
  (syntax-rules ()
    ((_ name type args)
     (define name
       (let ((proc (delay (pointer->procedure
                           type
                           (dynamic-func (symbol->string (quote name)) liblepton)
                           args))))
         (force proc))))))

(define-lff set_render_placeholders void '())
(define-lff colors_count size_t '())
(define-lff lepton_colormap_color_by_id '* (list '* size_t))
(define-lff lepton_colormap_disable_color void (list '* size_t))
(define-lff lepton_colormap_set_color void (list '* size_t uint8 uint8 uint8 uint8))
(define-lff print_colors_array '* '())
(define-lff s_clib_add_command '* '(* * *))
(define-lff s_clib_add_directory '* '(* *))
(define-lff s_clib_add_scm '* '(* * *))
(define-lff s_clib_init void '())
(define-lff s_toplevel_page_current '* (list '*))
;;; g_rc.c
(define-lff g_rc_parse void '(* * *))
;;; s_attrib.c
(define-lff s_attrib_uniq int (list '*))
(define-lff s_attrib_add_entry int (list '*))

;;; scheme_smob.c
(define-lff edascm_is_config int '(*))
(define-lff edascm_is_object int '(*))
(define-lff edascm_is_page int '(*))
(define-lff edascm_to_object '* '(*))
(define-lff edascm_from_object '* '(*))

;;; object.c
(define-lff lepton_object_get_color int '(*))
(define-lff lepton_object_set_color void (list '* int))
(define-lff lepton_object_get_id int '(*))
(define-lff lepton_object_get_parent '* '(*))
(define-lff lepton_object_get_selectable int '(*))
(define-lff lepton_object_set_selectable void (list '* int))
(define-lff lepton_object_get_type int '(*))

(define-lff lepton_object_is_arc int '(*))
(define-lff lepton_object_is_box int '(*))
(define-lff lepton_object_is_bus int '(*))
(define-lff lepton_object_is_circle int '(*))
(define-lff lepton_object_is_component int '(*))
(define-lff lepton_object_is_line int '(*))
(define-lff lepton_object_is_net int '(*))
(define-lff lepton_object_is_path int '(*))
(define-lff lepton_object_is_picture int '(*))
(define-lff lepton_object_is_pin int '(*))
(define-lff lepton_object_is_text int '(*))

(define-lff lepton_object_calculate_visible_bounds int (list '* int '* '* '* '*))
(define-lff lepton_object_copy '* '(*))
(define-lff lepton_object_emit_change_notify void '(*))
(define-lff lepton_object_emit_pre_change_notify void '(*))
(define-lff lepton_object_mirror void (list int int '*))
(define-lff lepton_object_rotate void (list int int int '*))
(define-lff lepton_object_translate void (list '* int int))

(define-lff lepton_object_page_set_changed void '(*))

(define-lff lepton_component_object_get_embedded int '(*))
(define-lff lepton_component_object_embed void '(*))
(define-lff lepton_component_object_unembed void '(*))

(define-lff lepton_picture_object_get_embedded int '(*))
(define-lff lepton_picture_object_embed void '(*))
(define-lff lepton_picture_object_unembed void '(*))

;;; Helper transformers between #<geda-object> smobs and C object
;;; pointers.
(define (geda-object->pointer smob)
  (or (false-if-exception (edascm_to_object (scm->pointer smob)))
      ;; Return NULL if the SMOB is not the #<geda-object> smob.
      %null-pointer))

(define (pointer->geda-object pointer)
  ;; Return #f if the pointer is wrong.
  (false-if-exception (pointer->scm (edascm_from_object pointer))))

;;; This syntax rule is intended for use in toplevel 'define' or
;;; 'let' forms in the functions where the check for wrong type of
;;; OBJECT is necessary.  The rule checks the object and, if it is
;;; not #<geda-object>, throws an error with the 'wrong-type-arg
;;; key reporting the function name and position POS of the
;;; OBJECT argument.  In short, the usage is as follows:
;;;   (define (myfunc object)
;;;     (define pointer (geda-object->pointer* object 1))
;;;     (function-body))
(define-syntax-rule (geda-object->pointer* object pos)
  (let ((pointer (geda-object->pointer object)))
    (if (null-pointer? pointer)
        (let ((proc-name
               ;; Provision agains Guile-2.0 that does not have the procedure.
               (if (defined? 'frame-procedure-name)
                   (frame-procedure-name (stack-ref (make-stack #t) 1))
                   '??)))
          (scm-error 'wrong-type-arg
                     proc-name
                     "Wrong type argument in position ~A: ~A"
                     (list pos object)
                     #f))
        pointer)))
